(* File absyn.ml *)
(* Amr Sabry *)

(* This is the output of the parser *)

type prim = 
    Minus
  | Times
  | Plus
  | Less
  | Leq
  | Eq
  | Add1
  | Booleanp
  | Car
  | Cdr
  | Cons
  | Integerp
  | Not
  | Nullp
  | Pairp
  | Procedurep
  | Sub1
  | Void
  | Zerop
  | Eqp
  | Setcar
  | Setcdr

let opArity op = 
  match op with 
    (Minus | Times | Plus) -> 2
  | (Less | Leq | Eq | Eqp) -> 2
  | (Cons | Setcar | Setcdr) -> 2
  | (Car | Cdr) -> 1
  | (Add1 | Sub1 | Not) -> 1
  | (Booleanp | Integerp | Nullp | Pairp | Procedurep | Zerop) -> 1
  | (Void) -> 0


let opName op = 
  match op with 
    Minus -> "-"
  | Times -> "*"
  | Plus -> "+"
  | Less -> "<"
  | Leq -> "<="
  | Eq -> "="
  | Add1 -> "add1"
  | Booleanp -> "boolean?"
  | Car -> "car"
  | Cdr -> "cdr"
  | Cons -> "cons"
  | Integerp -> "integer?"
  | Not -> "not"
  | Nullp -> "null?"
  | Pairp -> "pair?"
  | Procedurep -> "procedure?"
  | Sub1 -> "sub1"
  | Void -> "void"
  | Zerop -> "zero?"
  | Eqp -> "eq?"
  | Setcar -> "set-car!"
  | Setcdr -> "set-cdr!"

type const = 
    BoolC of bool
  | IntC of int

type datum = 
    ConstantD of const 
  | EmptyListD 

type exp = 
    ConstantE of const 
  | QuoteE of datum 
  | VarE of string
  | SetE of string * exp
  | OneifE of exp * exp
  | IfE of exp * exp * exp
  | BeginE of exp list
  | LetE of string list * exp list * exp 
  | LetrecE of string list * exp list * exp 
  | LambdaE of string list * exp 
  | PrimappE of prim * exp list
  | AppE of exp * exp list

(********************************************************************)
(* Boring pretty-printing routines *)

let print_const c = 
  match c with 
    BoolC b -> Format.print_string (if b then "#t" else "#f")
  | IntC i -> Format.print_int i

let rec print_datum d = 
  match d with
    ConstantD c -> print_const c
  | EmptyListD -> Format.print_string "()"

let rec print_exp e = 
  match e with 
    ConstantE c -> print_const c
  | QuoteE e -> begin 
                  Format.open_hbox(); 
                  Format.print_string "'"; 
                  print_datum e; 
                  Format.close_box()
                end
  | VarE e -> Format.print_string e
  | SetE (x,e) -> begin
                    Format.open_hvbox 4;
                    Format.print_string "(set! ";
                    Format.print_string x;
                    Format.print_space();
                    print_exp e; 
                    Format.print_string ")";
                    Format.close_box()
                  end
  | OneifE (e1,e2) -> begin
                        Format.open_hvbox 4;
                        Format.print_string "(if ";
                        print_exp e1; 
                        Format.print_space();
                        print_exp e2; 
                        Format.print_string ")";
                        Format.close_box()
                      end
  | IfE (e1,e2,e3) -> begin
                        Format.open_hvbox 4;
                        Format.print_string "(if ";
                        print_exp e1;
                        Format.print_space();
                        print_exp e2;
                        Format.print_space();
                        print_exp e3;
                        Format.print_string ")";
                        Format.close_box()
                      end
  | BeginE es -> begin
                   Format.open_hvbox 2;
                   Format.print_string "(begin";
                   Format.print_space();
                   Util.formatseq es print_exp Format.print_space;
                   Format.print_string ")";
                   Format.close_box()
                 end
  | LetE (ss,es,e) -> begin
                       Format.open_hvbox 2;
                       Format.print_string "(let (";
                       Format.open_hvbox 0;
                       Util.formatseq
                         (List.combine ss es) 
                         (fun (s,e) -> 
                           Format.open_hbox();
                           Format.print_string "(";
                           Format.print_string s;
                           Format.print_space();
                           print_exp e;
                           Format.print_string ")";
                           Format.close_box())
                       Format.print_space;
                       Format.close_box();
                       Format.print_string ")";
                       Format.print_space();
                       print_exp e;
                       Format.print_string ")";
                       Format.close_box()
                     end
  | LetrecE (ss,es,e) -> begin
                           Format.open_hvbox 2;
                           Format.print_string "(letrec (";
                           Format.open_hvbox 0;
                           Util.formatseq
                             (List.combine ss es) 
                             (fun (s,e) -> 
                                Format.open_hbox();
                                Format.print_string "(";
                                Format.print_string s;
                                Format.print_space();
                                print_exp e;
                                Format.print_string ")";
                                Format.close_box())
                           Format.print_space;
                           Format.close_box();
                           Format.print_string ")";
                           Format.print_space();
                           print_exp e;
                           Format.print_string ")";
                           Format.close_box()
                        end
  | LambdaE (ss,e) -> begin
                        Format.open_hvbox 2;
                        Format.print_string "(lambda (";
                        Format.open_hvbox 0;
                        Util.formatseq ss Format.print_string Format.print_space;
                        Format.close_box();
                        Format.print_string ")";
                        Format.print_space();
                        print_exp e;
                        Format.print_string ")";
                        Format.close_box()
                      end
  | PrimappE (p,es) -> begin 
                         Format.open_hvbox 2;
                         Format.print_string "(";
                         Format.print_string (opName p);
                         Format.print_space();
                         Util.formatseq es print_exp Format.print_space;
                         Format.print_string ")";
                         Format.close_box()
                       end 
  | AppE (e,es) -> begin
                     Format.open_hvbox 2;
                     Format.print_string "(";
                     print_exp e;
                     Format.print_space();
                     Util.formatseq es print_exp Format.print_space;
                     Format.print_string ")";
                     Format.close_box()
                   end

let print e = 
  Format.open_hbox();
  print_exp e;
  Format.close_box();
  Format.print_newline()

(********************************************************************)
(********************************************************************)
