(* John David Ratliff
   Monday 2:30-3:20 Lab
   jdratlif@cs.indiana.edu *)

(******************************************************************************)
(* This interpreter assumes that we use Ocaml's references to implement *)
(* the assignments of the language we are interpreting (Scheme) *) 

exception Error of string

(******************************************************************************)
(* Implementation of environments *)

module VarType = struct
  type t = string
  let compare = Pervasives.compare 
end 

module Env = Map.Make (VarType)

type rtv = 
    BoolV of bool 
  | IntV of int 
  | EmptyV 
  | VoidV 
  | PairV of rtv ref * rtv ref (* because Scheme pairs are assignable *)
  | ClosureV of string list * Absyn.exp * env 
and env = (rtv ref) Env.t    (* because Scheme variables are assignable *)

(******************************************************************************)
(* Interpreter *)

(******************************************************************************)
(* Some operations on run-time values: checking if a value is true and applying *)
(* primitive operations like addition, multiplication, etc *)

let true_val v = match v with
  BoolV false -> false
| _ -> true

let primApp p vs = match (p,vs) with

  (Absyn.Minus,[IntV i1; IntV i2]) -> IntV (i1-i2)
| (Absyn.Times,[IntV i1; IntV i2]) -> IntV (i1*i2)
| (Absyn.Plus, [IntV i1; IntV i2]) -> IntV (i1+i2)
| (Absyn.Less, [IntV i1; IntV i2]) -> if i1 < i2 then BoolV true else BoolV false
| (Absyn.Leq,  [IntV i1; IntV i2]) -> if i1 <= i2 then BoolV true else BoolV false
| (Absyn.Eq,   [IntV i1; IntV i2]) -> if i1 = i2 then BoolV true else BoolV false

| (Absyn.Add1, [IntV i]) -> IntV (i+1)
| (Absyn.Sub1, [IntV i]) -> IntV (i-1)
| (Absyn.Zerop,[IntV i]) -> if i = 0 then BoolV true else BoolV false

| (Absyn.Car,[PairV(v1,_)]) -> !v1
| (Absyn.Cdr,[PairV(_,v2)]) -> !v2
| (Absyn.Cons,[v1;v2]) -> PairV(ref v1,ref v2)

| (Absyn.Not,[v]) -> if true_val v then BoolV false else BoolV true

| (Absyn.Booleanp,[BoolV _]) -> BoolV true
| (Absyn.Booleanp,_) -> BoolV false

| (Absyn.Integerp,[IntV _]) -> BoolV true
| (Absyn.Integerp,_) -> BoolV false

| (Absyn.Nullp,[EmptyV]) -> BoolV true
| (Absyn.Nullp,[_]) -> BoolV false

| (Absyn.Pairp,[PairV _]) -> BoolV true
| (Absyn.Pairp,_) -> BoolV false

| (Absyn.Procedurep,[ClosureV _]) -> BoolV true
| (Absyn.Procedurep,_) -> BoolV false

| (Absyn.Void,[]) -> VoidV

| (Absyn.Setcar,[PairV(car,_); value]) -> car := value; VoidV
| (Absyn.Setcar,[_; value]) -> raise (Error "Error: in (set-car! ls value) ls must be a pair.")

| (Absyn.Setcdr,[PairV(_,cdr); value]) -> cdr := value; VoidV
| (Absyn.Setcdr,[_; value]) -> raise (Error "Error: in (set-cdr! ls value) ls must be a pair.")

| (Absyn.Eqp,[v1; v2]) -> BoolV (v1 == v2)

| _ -> raise (Error "Illegal primitive application")

(******************************************************************************)
(* Here starts the code for the interpreter: the main function is "eval" *)
(* This interpreter is for static-scope: it is missing three clauses that *)
(* are to be completed *)

let eval_const c = match c with
  Absyn.BoolC b -> BoolV b
| Absyn.IntC i -> IntV i

let eval_datum d = match d with
  Absyn.ConstantD c -> eval_const c
| Absyn.EmptyListD -> EmptyV

(*******************************************************************)
(* Static scope *)

let rec eval_exp e env = match e with
  Absyn.ConstantE c -> eval_const c 
| Absyn.QuoteE d -> eval_datum d
| Absyn.VarE s -> ! (Env.find s env)
| Absyn.OneifE (e1,e2) -> if true_val (eval_exp e1 env) then eval_exp e2 env else VoidV
| Absyn.IfE (e1,e2,e3) -> if true_val (eval_exp e1 env) then eval_exp e2 env else eval_exp e3 env
| Absyn.BeginE es -> Util.last (eval_exps es env)
| Absyn.LetE (ss,es,e) -> 
    let vs = eval_exps es env in
    let newenv = List.fold_left2 (fun env s v -> Env.add s (ref v) env) env ss vs in
    eval_exp e newenv
| Absyn.PrimappE (p,es) -> let vs = eval_exps es env in primApp p vs
| Absyn.LetrecE (ss,es,e) -> 
    let renv = List.fold_left (fun env s -> Env.add s (ref VoidV) env) env ss in
    let vs = eval_exps es renv in
    List.iter2 (fun s v -> let r = Env.find s renv in r := v) ss vs;
    eval_exp e renv
| Absyn.LambdaE (ss,e) -> ClosureV (ss,e,env)
| Absyn.SetE (var,exp) -> 
    let value = eval_exp exp env in
    let ptr = Env.find var env in
    ptr := value;
    VoidV
| Absyn.AppE (e,es) -> 
    let f = eval_exp e env in
    let args = eval_exps es env in
    match f with
      ClosureV(ss,body,fenv) -> 
        let benv = List.fold_left2 (fun env s v -> Env.add s (ref v) env) fenv ss args in
        eval_exp body benv
    | _ -> raise (Error "Trying to apply a non-function")

and eval_exps es env = List.map (fun e -> eval_exp e env) es

let eval e = eval_exp e Env.empty

(********************************************************************)
(* Pretty-printing *)

let rec print_val v = match v with
    BoolV b -> Format.print_string (if b then "#t" else "#f")
  | IntV i -> Format.print_int i
  | EmptyV -> Format.print_string "()"
  | VoidV -> Format.print_string "VOID"
  | PairV (r1,r2) -> (* won't work for circular lists *)
      Format.open_hvbox 2;
      Format.print_string "(";
      print_val (!r1); 
      Format.print_string " . ";
      print_val (!r2);
      Format.print_string ")";
      Format.close_box()
  | ClosureV _ -> Format.print_string "<procedure>"

let print v = 
  Format.open_hbox();
  print_val v;
  Format.close_box();
  Format.print_newline()

(********************************************************************)
(********************************************************************)
