(* John David Ratliff
   Monday 2:30-3:20 Lab
   jdratlif@cs.indiana.edu *)

(******************************************************************************)
(* This interpreter should use store-passing style to avoid using Ocaml's references *)
(* See Exercise 3.48 in the book *)

exception Error of string

(******************************************************************************)
(* Environments map variables to locations *)

module VarType = struct
  type t = string
  let compare = Pervasives.compare 
end 

module Env = Map.Make (VarType)

type location = int 

type rtv = 
    BoolV of bool 
  | IntV of int 
  | EmptyV 
  | VoidV 
  | PairV of location * location (* pointer to the car/cdr in the store *)
  | ClosureV of string list * Absyn.exp * env 
and env = location Env.t  (* environment maps variables to locations *)

(* Stores map locations to values *)

module LocType = struct
  type t = location
  let compare = Pervasives.compare
end

module StoreM = Map.Make (LocType)

module type STORE = sig
  type store 
  val empty : store
  val alloc : rtv -> store -> (location * store)
  val allocN : int -> store -> (location list * store)
  val update : location -> rtv -> store -> store
  val find : location -> store -> rtv
end 

module Store : STORE = struct
  type location = int 
  (* the first component is the next free location *)
  (* the second component is a map from locations to values *)
  type store = location * rtv StoreM.t
  
  (* returns a new store *)
  let empty = (0, StoreM.empty)

  (* Allocate and initialize a pointer to v, return the pointer and the new store *)
  let alloc v (fp,sm) = 
    let ns = (fp+1, StoreM.add fp v sm) in
    (fp, ns)

  (* Allocate n pointers initialized to VoidV *)
  (* returns a list of the pointers and the new store *)
  let rec allocN n s = 
    if n = 0 
    then ([],s)
    else let (p1,s1) = alloc VoidV s in
         let (ps,sf) = allocN (n-1) s1 in
         (p1::ps, sf)

  (* Updates an existing location in the store; returns the new store *)
  let update loc v (fp,sm) = (fp, StoreM.add loc v sm)

  (* Finds the value associated with a location *)
  let find loc (_,sm) = StoreM.find loc sm
end 

(******************************************************************************)
(* TODO *)

let true_val v = match v with
  BoolV false -> false
| _ -> true

let primApp p vs store = match (p,vs) with

  (Absyn.Minus,[IntV i1; IntV i2]) -> (IntV (i1-i2), store)
| (Absyn.Times,[IntV i1; IntV i2]) -> (IntV (i1*i2), store)
| (Absyn.Plus, [IntV i1; IntV i2]) -> (IntV (i1+i2), store)
| (Absyn.Less, [IntV i1; IntV i2]) -> if i1 < i2 then (BoolV true, store) else (BoolV false, store)
| (Absyn.Leq,  [IntV i1; IntV i2]) -> if i1 <= i2 then (BoolV true, store) else (BoolV false, store)
| (Absyn.Eq,   [IntV i1; IntV i2]) -> if i1 = i2 then (BoolV true, store) else (BoolV false, store)

| (Absyn.Add1, [IntV i]) -> (IntV (i+1), store)
| (Absyn.Sub1, [IntV i]) -> (IntV (i-1), store)
| (Absyn.Zerop,[IntV i]) -> if i = 0 then (BoolV true, store) else (BoolV false, store)

| (Absyn.Car,[PairV(v1,_)]) -> (Store.find v1 store, store)
| (Absyn.Cdr,[PairV(_,v2)]) -> (Store.find v2 store, store)
| (Absyn.Cons,[v1;v2]) -> 
    let (ptr1,newstore1) = Store.alloc v1 store in
    let (ptr2,newstore2) = Store.alloc v2 newstore1 in
    (PairV(ptr1,ptr2), newstore2)

| (Absyn.Not,[v]) -> if true_val v then (BoolV false, store) else (BoolV true, store)

| (Absyn.Booleanp,[BoolV _]) -> (BoolV true, store)
| (Absyn.Booleanp,_) -> (BoolV false, store)

| (Absyn.Integerp,[IntV _]) -> (BoolV true, store)
| (Absyn.Integerp,_) -> (BoolV false, store)

| (Absyn.Nullp,[EmptyV]) -> (BoolV true, store)
| (Absyn.Nullp,[_]) -> (BoolV false, store)

| (Absyn.Pairp,[PairV _]) -> (BoolV true, store)
| (Absyn.Pairp,_) -> (BoolV false, store)

| (Absyn.Procedurep,[ClosureV _]) -> (BoolV true, store)
| (Absyn.Procedurep,_) -> (BoolV false, store)

| (Absyn.Void,[]) -> (VoidV, store)

| (Absyn.Setcar,[PairV(car,_); value]) -> 
    let newstore = Store.update car value store in
    (VoidV, newstore)
| (Absyn.Setcar,[_; value]) -> raise (Error "Error: in (set-car! ls value) ls must be a pair.")

| (Absyn.Setcdr,[PairV(_,cdr); value]) -> 
    let newstore = Store.update cdr value store in
    (VoidV, newstore)
| (Absyn.Setcdr,[_; value]) -> raise (Error "Error: in (set-cdr! ls value) ls must be a pair.")

| (Absyn.Eqp,[v1; v2]) -> (BoolV (v1 == v2), store)

| _ -> raise (Error "Illegal primitive application")

(******************************************************************************)
(* Here starts the code for the interpreter: the main function is "eval" *)
(* This interpreter is for static-scope: it is missing three clauses that *)
(* are to be completed *)

let eval_const c = match c with
  Absyn.BoolC b -> BoolV b
| Absyn.IntC i -> IntV i

let eval_datum d = match d with
  Absyn.ConstantD c -> eval_const c
| Absyn.EmptyListD -> EmptyV

(*******************************************************************)
(* Static scope *)

let rec eval_exp e env store = match e with
  Absyn.ConstantE c -> (eval_const c, store)
| Absyn.QuoteE d -> (eval_datum d, store)
| Absyn.VarE s -> (Store.find (Env.find s env) store, store)
| Absyn.OneifE (e1,e2) -> 
    let (value,_) = eval_exp e1 env store in
    if true_val value then eval_exp e2 env store else (VoidV,store)
| Absyn.IfE (e1,e2,e3) -> 
    let (value,_) = eval_exp e1 env store in
    if true_val value then eval_exp e2 env store else eval_exp e3 env store
| Absyn.BeginE es -> 
    let (value,newstore) = Util.last (eval_exps es env store) in
    (value,newstore)
| Absyn.PrimappE (p,es) -> 
    let (vs,stores) = List.split (eval_exps es env store) in
    let newstore = if (List.length stores) > 0 then Util.last stores else store in
    primApp p vs newstore
| Absyn.LambdaE (ss,e) -> (ClosureV (ss,e,env), store)
| Absyn.SetE (var,exp) -> 
    let (value,_) = eval_exp exp env store in
    let ptr = Env.find var env in
    let newstore = Store.update ptr value store in
    (VoidV, newstore)
| Absyn.LetE (ss,es,e) ->
    let (values,stores) = List.split (eval_exps es env store) in
    let newstore = if (List.length stores) > 0 then Util.last stores else store in
    let (ptrs,newstore2) = Store.allocN (List.length values) newstore in
    let rec bangbang ptrlst vs cs = match ptrlst with
        [] -> cs |
        (loc::cdr) -> bangbang cdr (List.tl vs) (Store.update loc (List.hd vs) cs)
      in
    let finalstore = bangbang ptrs values newstore2 in
    let newenv = List.fold_left2 (fun env s v -> Env.add s v env) env ss ptrs in
      eval_exp e newenv finalstore
| Absyn.LetrecE (ss,es,e) ->
    let (ptrs,rstore) = Store.allocN (List.length ss) store in
    let renv = List.fold_left2 (fun env s v -> Env.add s v env) env ss ptrs in
    let (values,stores) = List.split (eval_exps es renv rstore) in
    let newstore = if (List.length values) > 0 then Util.last stores else rstore in
    let rec bangbang ptrlst vs cs = match ptrlst with
        [] -> cs |
        (loc::cdr) -> bangbang cdr (List.tl vs) (Store.update loc (List.hd vs) cs)
      in
    let finalstore = bangbang ptrs values newstore in
      eval_exp e renv finalstore
| Absyn.AppE (e,es) ->
    let f = eval_exp e env store in
    let (args,stores) = List.split (eval_exps es env store) in
    let newstore = if (List.length args) = 0 then Util.last stores else store in
    match f with
      (ClosureV(ss,body,fenv), ignore) ->
          let (ptrs,newstore2) = Store.allocN (List.length args) newstore in
	  let rec bangbang ptrlst vs cs = match ptrlst with
	      [] -> cs |
	      (loc::cdr) -> bangbang cdr (List.tl vs) (Store.update loc (List.hd vs) cs)
	    in
	  let finalstore = bangbang ptrs args newstore2 in
	  let newenv = List.fold_left2 (fun env s v -> Env.add s v env) fenv ss ptrs in
            eval_exp body newenv finalstore |
      _ -> raise (Error "Not a procedure")
and eval_exps es env store = 
    let rec helper exps newstore = match exps with
      [] -> [] |
      (e::cdr) -> let (value,newstore2) = eval_exp e env newstore in (value,newstore2) :: helper cdr newstore2
    in
    helper es store

let eval e = eval_exp e Env.empty Store.empty

(********************************************************************)
(* Pretty-printing *)

let rec print_val v s = match v with
    BoolV b -> Format.print_string (if b then "#t" else "#f")
  | IntV i -> Format.print_int i
  | EmptyV -> Format.print_string "()"
  | VoidV -> Format.print_string "VOID"
  | PairV (r1,r2) -> 
      Format.open_hvbox 2;
      Format.print_string "(";
      print_val (Store.find r1 s) s; 
      Format.print_string " . ";
      print_val (Store.find r2 s) s;
      Format.print_string ")";
      Format.close_box()
  | ClosureV _ -> Format.print_string "<procedure>"

let print v s = 
  Format.open_hbox();
  print_val v s;
  Format.close_box();
  Format.print_newline()

(********************************************************************)
(********************************************************************)
